/**
 * \file mcql/core/peptidemeasurements.h
 * \date 26/09/2024
 * \author Olivier Langella
 * \brief peptide extracted measures model in masschroqlite
 */

/*******************************************************************************
 * Copyright (c) 2024 Olivier Langella
 *<Olivier.Langella@universite-paris-saclay.fr>.
 *
 * This file is part of MassChroQ.
 *
 *     MassChroQ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MassChroQ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MassChroQ.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

#pragma once


#include <cstdint>
#include "quantificationmethod.h"
#include <pappsomspp/msrun/msrunid.h>
#include <pappsomspp/msrun/xiccoord/xiccoord.h>
#include "../../mcq_types.h"

#include <pappsomspp/peptide/peptidenaturalisotopelist.h>


namespace mcql
{

class PeptideObservation;

typedef std::shared_ptr<PeptideObservation> PeptideObservationSp;


class PeptideMeasurements;

typedef std::shared_ptr<PeptideMeasurements> PeptideMeasurementsSp;

class MsRunGroup;

/**
 * group together a peptide and measurements made on this peptide
 */
class PeptideMeasurements
{
  public:
  /**
   * Default constructor
   */
  PeptideMeasurements(const mcql::PeptideObservationSp &peptide_observations);

  /**
   * Copy constructor
   *
   * @param other TODO
   */
  PeptideMeasurements(const PeptideMeasurements &other);

  /**
   * Destructor
   */
  virtual ~PeptideMeasurements();

  void prepareMeasurements(const pappso::MsRunId &targeted_msrun,
                           const mcql::MsRunGroup *msrun_group_p,
                           const QuantificationMethodSp &quantification_method);

  void
  pushBackXicCoordList(std::vector<pappso::XicCoordSPtr> &xic_coord_list) const;

  /** the same peptide can be measured multiple times :
   * - for each charge state
   * - for each isotopologue
   * One measurement contains :
   * - a transient chromatogram extraction
   * - extracted XIC and coordinates
   * - peak quality : how the detected peak in the chromatogram was chosen
   *
   */
  struct Measurement
  {
    std::uint8_t m_charge;
    pappso::XicCoordSPtr msp_xicCoord;
    PeakQualityCategory m_peakQualityCategory = PeakQualityCategory::nomatch;
    pappso::TracePeakCstSPtr m_tracePeakSp;
    pappso::PeptideNaturalIsotopeAverageSp m_peptideNaturalIsotopeAverageSp;
  };

  void detectQuantifyPeaks(const QuantificationMethodSp &quantification_method);

  const std::vector<Measurement> &getMeasurementList() const;
  const mcql::PeptideObservationSp &getPeptideObservationSp() const;

  void flushXics();


  /** @brief clear the measurement vector
   */
  void clear();

  private:
  mcql::PeptideObservationSp msp_peptideObservation;

  std::vector<Measurement> m_measurementList;
};
} // namespace mcql
