/**
 * \file mcql/core/quantificationmethod.cpp
 * \date 24/10/2024
 * \author Olivier Langella
 * \brief store parameters related to peak detection and quantification
 */

/*******************************************************************************
 * Copyright (c) 2024 Olivier Langella
 *<Olivier.Langella@universite-paris-saclay.fr>.
 *
 * This file is part of MassChroQ.
 *
 *     MassChroQ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MassChroQ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MassChroQ.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

#include "quantificationmethod.h"
#include "../utils.h"
#include <pappsomspp/pappsoexception.h>
#include <pappsomspp/exception/exceptionnotfound.h>
#include <pappsomspp/processing/detection/tracedetectionzivy.h>

mcql::QuantificationMethod::QuantificationMethod(const QString &id) : m_id(id)
{
}

mcql::QuantificationMethod::~QuantificationMethod()
{
}

bool
mcql::QuantificationMethod::getMatchBetweenRun() const
{
  return m_matchBetweenRun;
}

double
mcql::QuantificationMethod::getIsotopeMinimumRatio() const
{
  return m_isotopeMinimumRatio;
}

void
mcql::QuantificationMethod::addXicFilter(
  const pappso::FilterNameInterfaceSPtr &filter)
{
  if(m_xicFilterSuite.get() == nullptr)
    {
      m_xicFilterSuite =
        std::make_shared<pappso::FilterSuiteString>(filter.get()->toString());
    }
  else
    {
      m_xicFilterSuite.get()->addFilter(filter);
    }
}

pappso::PrecisionPtr
mcql::QuantificationMethod::getXicExtractionMeanPrecisionPtr() const
{

  if(mp_xicExtractionLowerPrecisionPtr->unit() == pappso::PrecisionUnit::ppm)
    {
      return pappso::PrecisionFactory::getPpmInstance(
        (mp_xicExtractionLowerPrecisionPtr->getNominal() +
         mp_xicExtractionUppersPrecisionPtr->getNominal()) /
        2);
    }
  return pappso::PrecisionFactory::getDaltonInstance(
    (mp_xicExtractionLowerPrecisionPtr->getNominal() +
     mp_xicExtractionUppersPrecisionPtr->getNominal()) /
    2);
}

pappso::XicExtractMethod
mcql::QuantificationMethod::getXicExtractMethod() const
{
  return m_xicExtractMethod;
}

void
mcql::QuantificationMethod::setTraceDetectionInterfaceCstSPtr(
  const pappso::TraceDetectionInterfaceCstSPtr &detection)
{
  mcsp_traceDetectionInterfaceCstSPtr = detection;
}

void
mcql::QuantificationMethod::setXicExtractionLowerPrecisionPtr(
  pappso::PrecisionPtr precision)
{
  mp_xicExtractionLowerPrecisionPtr = precision;
}

void
mcql::QuantificationMethod::setXicExtractionUpperPrecisionPtr(
  pappso::PrecisionPtr precision)
{
  mp_xicExtractionUppersPrecisionPtr = precision;
}

void
mcql::QuantificationMethod::setXicExtractMethod(pappso::XicExtractMethod method)
{
  m_xicExtractMethod = method;
}

const pappso::FilterNameInterfaceSPtr
mcql::QuantificationMethod::getXicFilter() const
{
  return m_xicFilterSuite;
}

const pappso::MzRange
mcql::QuantificationMethod::getXicExtractionMzRange(double mz) const
{

  return pappso::MzRange(
    mz, mp_xicExtractionLowerPrecisionPtr, mp_xicExtractionUppersPrecisionPtr);
}

const pappso::TraceDetectionInterfaceCstSPtr &
mcql::QuantificationMethod::getTraceDetectionInterfaceCstSPtr() const
{
  return mcsp_traceDetectionInterfaceCstSPtr;
}

pappso::ProjectParameters
mcql::QuantificationMethod::getProjectParameters() const
{
  pappso::ProjectParameters parameters;

  pappso::ProjectParam project_param(
    {pappso::ProjectParamCategory::quantification, "", QVariant()});
  // xic_type="max"
  project_param.name = "mcq_xic_extraction_type";
  project_param.value.setValue(mcql::Utils::enumToString(m_xicExtractMethod));
  parameters.setProjectParam(project_param);

  project_param.name = "mcq_mbr";
  project_param.value.setValue(getMatchBetweenRun());
  parameters.setProjectParam(project_param);
  
  project_param.name = "mcq_isotope_minimum_ratio";
  project_param.value.setValue(getIsotopeMinimumRatio());
  parameters.setProjectParam(project_param);


  if(mp_xicExtractionLowerPrecisionPtr->unit() == pappso::PrecisionUnit::ppm)
    {

      //<!--For XIC extraction on Da use: mz_range-->
      //<ppm_range min="10" max="10"/>

      project_param.name = "mcq_xic_ppm_range_min";
      project_param.value.setValue(
        mp_xicExtractionLowerPrecisionPtr->getNominal());
      parameters.setProjectParam(project_param);

      project_param.name = "mcq_xic_ppm_range_max";
      project_param.value.setValue(
        mp_xicExtractionUppersPrecisionPtr->getNominal());
      parameters.setProjectParam(project_param);
    }
  if(mp_xicExtractionLowerPrecisionPtr->unit() == pappso::PrecisionUnit::dalton)
    {

      //<!--For XIC extraction on Da use: mz_range-->
      //<ppm_range min="10" max="10"/>

      project_param.name = "mcq_xic_mz_range_min";
      project_param.value.setValue(
        mp_xicExtractionLowerPrecisionPtr->getNominal());
      parameters.setProjectParam(project_param);

      project_param.name = "mcq_xic_mz_range_max";
      project_param.value.setValue(
        mp_xicExtractionUppersPrecisionPtr->getNominal());
      parameters.setProjectParam(project_param);
    }

  project_param.name = "mcq_xic_pre_filter";
  if(m_xicFilterSuite.get() != nullptr)
    project_param.value.setValue(m_xicFilterSuite.get()->toString());
  else
    project_param.value.setValue("");
  parameters.setProjectParam(project_param);


  const pappso::TraceDetectionZivy *detection_zivy =
    dynamic_cast<const pappso::TraceDetectionZivy *>(
      mcsp_traceDetectionInterfaceCstSPtr.get());

  if(detection_zivy != nullptr)
    {
      project_param.name = "mcq_detection_zivy";
      project_param.value.setValue(
        QString("%1 %2 %3 %4 %5")
          .arg(detection_zivy->getSmoothingHalfEdgeWindows())
          .arg(detection_zivy->getMinMaxHalfEdgeWindows())
          .arg(detection_zivy->getMaxMinHalfEdgeWindows())
          .arg(detection_zivy->getDetectionThresholdOnMinmax())
          .arg(detection_zivy->getDetectionThresholdOnMaxmin()));
      parameters.setProjectParam(project_param);
    }
  return parameters;
}

void
mcql::QuantificationMethod::setJsonObject(
  const QJsonObject &quantification_method)
{
  m_matchBetweenRun = quantification_method.value("match_between_run").toBool();
  m_isotopeMinimumRatio =
    quantification_method.value("isotope_minimum_ratio").toDouble();
  QString filter_str = quantification_method.value("pre_filter").toString();

  qDebug() << filter_str;
  if(!filter_str.isEmpty())
    {
      pappso::FilterNameInterfaceSPtr filter_sp =
        std::make_shared<pappso::FilterSuiteString>(filter_str);
      addXicFilter(filter_sp);
      qDebug() << filter_str;
    }
  QJsonObject extraction = quantification_method.value("extraction").toObject();
  QString integration    = extraction.value("integration").toString();
  if(integration == "sum")
    {
      setXicExtractMethod(pappso::XicExtractMethod::sum);
    }
  else if(integration == "max")
    {
      setXicExtractMethod(pappso::XicExtractMethod::max);
    }
  else
    {
      throw pappso::ExceptionNotFound(
        QObject::tr("missing "
                    "masschroq_methods>quantification_method>extraction>"
                    "integration %1 value")
          .arg(integration));
    }

  QJsonObject precision = extraction.value("precision").toObject();
  qDebug() << precision.value("down").toDouble();
  qDebug() << precision.value("up").toDouble();

  if(precision.value("unit").toString() == "dalton")
    {
      setXicExtractionLowerPrecisionPtr(
        pappso::PrecisionFactory::getDaltonInstance(
          precision.value("down").toDouble()));
      setXicExtractionUpperPrecisionPtr(
        pappso::PrecisionFactory::getDaltonInstance(
          precision.value("up").toDouble()));
    }
  else if(precision.value("unit").toString() == "ppm")
    {
      setXicExtractionLowerPrecisionPtr(
        pappso::PrecisionFactory::getPpmInstance(
          precision.value("down").toDouble()));
      setXicExtractionUpperPrecisionPtr(
        pappso::PrecisionFactory::getPpmInstance(
          precision.value("up").toDouble()));
    }
  else
    {
      throw pappso::ExceptionNotFound(QObject::tr(
        "missing "
        "masschroq_methods>quantification_method>extraction>precision>"
        "unit"));
    }


  QJsonObject detection = quantification_method.value("detection").toObject();
  if(detection.value("type").toString() == "zivy")
    {
      std::shared_ptr<pappso::TraceDetectionZivy> sp_detection_zivy =
        std::make_shared<pappso::TraceDetectionZivy>(
          detection.value("meanfilter").toInt(),
          detection.value("minmax").toInt(),
          detection.value("maxmin").toInt(),
          detection.value("threshold_on_max").toInt(),
          detection.value("threshold_on_min").toInt());
      setTraceDetectionInterfaceCstSPtr(sp_detection_zivy);
    }
  else
    {

      throw pappso::ExceptionNotFound(QObject::tr(
        "only masschroq_methods>quantification_method>detection>type == zivy "
        "allowed"));
    }
}

pappso::PrecisionPtr
mcql::QuantificationMethod::getXicExtractionLowerPrecisionPtr() const
{
  return mp_xicExtractionLowerPrecisionPtr;
}

pappso::PrecisionPtr
mcql::QuantificationMethod::getXicExtractionUppersPrecisionPtr() const
{
  return mp_xicExtractionUppersPrecisionPtr;
}
