/**
 * \file mcql/core/quantificationmethod.h
 * \date 24/10/2024
 * \author Olivier Langella
 * \brief store parameters related to peak detection and quantification
 */

/*******************************************************************************
 * Copyright (c) 2024 Olivier Langella
 *<Olivier.Langella@universite-paris-saclay.fr>.
 *
 * This file is part of MassChroQ.
 *
 *     MassChroQ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MassChroQ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MassChroQ.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

#pragma once

#include <memory>
#include <QString>
#include <pappsomspp/types.h>
#include <pappsomspp/mzrange.h>
#include <pappsomspp/processing/filters/filtersuitestring.h>
#include <pappsomspp/processing/detection/tracedetectioninterface.h>
#include <pappsomspp/processing/project/projectparameters.h>
#include <QJsonObject>

namespace mcql
{

class QuantificationMethod;

typedef std::shared_ptr<QuantificationMethod> QuantificationMethodSp;

/**
 * @todo write docs
 */
class QuantificationMethod
{
  public:
  /**
   * Default constructor
   */
  QuantificationMethod(const QString &id);

  /**
   * Destructor
   */
  virtual ~QuantificationMethod();


  void setXicExtractMethod(pappso::XicExtractMethod method);

  pappso::XicExtractMethod getXicExtractMethod() const;

  void setXicExtractionLowerPrecisionPtr(pappso::PrecisionPtr precision);

  void setXicExtractionUpperPrecisionPtr(pappso::PrecisionPtr precision);

  pappso::PrecisionPtr getXicExtractionMeanPrecisionPtr() const;

  void addXicFilter(const pappso::FilterNameInterfaceSPtr &filter);

  void setTraceDetectionInterfaceCstSPtr(
    const pappso::TraceDetectionInterfaceCstSPtr &detection);


  const pappso::TraceDetectionInterfaceCstSPtr &
  getTraceDetectionInterfaceCstSPtr() const;

  const pappso::FilterNameInterfaceSPtr getXicFilter() const;


  const pappso::MzRange getXicExtractionMzRange(double mz) const;
  pappso::ProjectParameters getProjectParameters() const;


  void setJsonObject(const QJsonObject &json_object);

  pappso::PrecisionPtr getXicExtractionLowerPrecisionPtr() const;
  pappso::PrecisionPtr getXicExtractionUppersPrecisionPtr() const;

  bool getMatchBetweenRun() const;

  double getIsotopeMinimumRatio() const;

  private:
  const QString m_id;
  pappso::XicExtractMethod m_xicExtractMethod = pappso::XicExtractMethod::max;

  pappso::PrecisionPtr mp_xicExtractionLowerPrecisionPtr  = nullptr;
  pappso::PrecisionPtr mp_xicExtractionUppersPrecisionPtr = nullptr;


  /// the xic filters
  pappso::FilterSuiteStringSPtr m_xicFilterSuite;


  /// the peak detection method for this quantification
  pappso::TraceDetectionInterfaceCstSPtr mcsp_traceDetectionInterfaceCstSPtr;

  bool m_matchBetweenRun = true;
  /** @brief the minimum percentage of theoretical intensity of the isotope
   * pattern to compute
   */
  double m_isotopeMinimumRatio = 0.8;
};
} // namespace mcql
